import logging
from pydantic import BaseModel

from . import (
    fusionjson,
    fusionzip,
    fusioncsv,
    empalccsv,
    empalcxlsx,
)

logger = logging.getLogger(__name__)


def process(
    fn: str,
    encoding: str = "utf-8",
    timezone: str = "localtime",
    parameters: BaseModel = None,
) -> tuple[list, dict, bool]:
    """
    Unified chromatographic data parser.

    Parameters
    ----------
    fn
        The file containing the trace(s) to parse.

    encoding
        Encoding of ``fn``, by default "utf-8".

    timezone
        A string description of the timezone. Default is "localtime".

    parameters
        Parameters for :class:`~dgbowl_schemas.yadg_dataschema.dataschema_4_2.step.ChromData`.

    Returns
    -------
    (data, metadata, fulldate) : tuple[list, dict, bool]
        Tuple containing the timesteps, metadata, and full date tag. All currently
        supported file formats return full date.
    """
    if parameters.filetype == "fusion.json":
        data, meta, fulldate = fusionjson.process(fn, encoding, timezone)
    elif parameters.filetype == "fusion.zip":
        data, meta, fulldate = fusionzip.process(fn, encoding, timezone)
    elif parameters.filetype == "fusion.csv":
        data, meta, fulldate = fusioncsv.process(fn, encoding, timezone)
    elif parameters.filetype == "empalc.csv":
        data, meta, fulldate = empalccsv.process(fn, encoding, timezone)
    elif parameters.filetype == "empalc.xlsx":
        data, meta, fulldate = empalcxlsx.process(fn, encoding, timezone)

    return data, meta, fulldate



"""
This parser handles the reading and processing of flow controller or flow meter 
data. 

Usage
`````
Available since ``yadg-4.0``. The parser supports the following parameters:

.. _yadg.parsers.flowdata.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_4_2.step.FlowData.Params

.. _yadg.parsers.flowdata.formats:

Formats
```````
The ``filetypes`` currently supported by the parser are:

  - DryCal log file text output (``drycal.txt``),
    see :mod:`~yadg.parsers.flowdata.drycal`
  - DryCal log file tabulated output (``drycal.csv``),
    see :mod:`~yadg.parsers.flowdata.drycal`
  - DryCal log file document file (``drycal.rtf``),
    see :mod:`~yadg.parsers.flowdata.drycal`

.. _yadg.parsers.flowdata.provides:

Provides
````````
The parser is used to extract all tabular data in the input file. This parser processes 
additional calibration information analogously to :mod:`~yadg.parsers.basiccsv`. 
        
.. admonition:: DEPRECATED in ``yadg-4.2``

  The processing of calibration information has been deprecated in ``yadg-4.2``
  and will stop working in ``yadg-5.0``.

"""
from .main import process



"""
**fusioncsv**: Processing Inficon Fusion csv export format (csv).
------------------------------------------------------------------

This is a tabulated format, including the concentrations, mole fractions, peak 
areas, and retention times. The latter is ignored by this parser.

.. warning::

    As also mentioned in the ``csv`` files themselves, the use of this filetype
    is discouraged, and the ``json`` files (or a zipped archive of them) should
    be parsed instead.


Exposed metadata:
`````````````````

.. code-block:: yaml

    params:
      method:   !!str
      username: None
      version:  None
      datafile: None

.. codeauthor:: Peter Kraus
"""
import logging
from ...dgutils.dateutils import str_to_uts
from uncertainties.core import str_to_number_with_uncert as tuple_fromstr

logger = logging.getLogger(__name__)

_headers = {
    "Concentration": ["concentration", "%"],
    "NormalizedConcentration": ["xout", "%"],
    "Area": ["area", " "],
    "RT(s)": ["retention time", "s"],
}


def process(fn: str, encoding: str, timezone: str) -> tuple[list, dict]:
    """
    Fusion csv export format.

    Multiple chromatograms per file, with multiple detectors.

    Parameters
    ----------
    fn
        Filename to process.

    encoding
        Encoding used to open the file.

    timezone
        Timezone information. This should be ``"localtime"``.

    Returns
    -------
    ([chrom], metadata, fulldate): tuple[list, dict, bool]
        Standard timesteps, metadata, and date tuple.
    """

    with open(fn, "r", encoding=encoding, errors="ignore") as infile:
        lines = infile.readlines()

    data = []
    for line in lines[3:]:
        if "SampleName" in line:
            header = [i.strip() for i in line.split(",")]
            sni = header.index("SampleName")
            method = header[0]
            for ii, i in enumerate(header):
                if i == "":
                    header[ii] = header[ii - 1]
        elif "Detectors" in line:
            detectors = [i.replace('"', "").strip() for i in line.split(",")]
            for ii, i in enumerate(detectors):
                if i == "":
                    detectors[ii] = detectors[ii - 1]
        elif "Time" in line:
            samples = [i.replace('"', "").strip() for i in line.split(",")]
            time = samples[0]
            if time == "Time (GMT 120 mins)":
                offset = "+02:00"
            elif time == "Time (GMT 60 mins)":
                offset = "+01:00"
            else:
                logger.error("offset '%s' not understood", time)
                offset = "+00:00"
        elif "% RSD" in line:
            continue
        else:
            items = line.split(",")
            point = {
                "concentration": {},
                "xout": {},
                "area": {},
                "retention time": {},
                "sampleid": items[sni],
            }
            uts = str_to_uts(f"{items[0]}{offset}", timezone=timezone)
            for ii, i in enumerate(items[2:]):
                ii += 2
                h = _headers.get(header[ii], None)
                chem = samples[ii]
                if h is None:
                    continue
                n, s = tuple_fromstr(i)
                point[h[0]][chem] = {
                    "n": n,
                    "s": s,
                    "u": h[1],
                }
            data.append({"uts": uts, "raw": point, "fn": fn})
    return data, {"params": {"method": method}}, True



"""
**empalccsv**: Processing Empa's online LC exported data (csv)
--------------------------------------------------------------

This is a structured format produced by the export from Agilent's Online LC device
at Empa. It contains three sections:

  - metadata section,
  - table containing sampling information,
  - table containing analysed chromatography data.


Exposed metadata:
`````````````````

.. code-block:: yaml

    params:
      method:   !!str
      username: !!str
      version:  !!int
      datafile: !!str

.. codeauthor:: Peter Kraus
"""
import logging
import datetime
from uncertainties.core import str_to_number_with_uncert as tuple_fromstr

logger = logging.getLogger(__name__)


def process(fn: str, encoding: str, timezone: str) -> tuple[list, dict]:
    """
    Fusion csv export format.

    Multiple chromatograms per file, with multiple detectors.

    Parameters
    ----------
    fn
        Filename to process.

    encoding
        Encoding used to open the file.

    timezone
        Timezone information. This should be ``"localtime"``.

    Returns
    -------
    ([chrom], metadata, fulldate): tuple[list, dict, bool]
        Standard timesteps, metadata, and date tuple.
    """

    with open(fn, "r", encoding=encoding, errors="ignore") as infile:
        lines = infile.readlines()

    metadata = {}
    while len(lines) > 0:
        line = lines.pop(0)
        if len(lines) == 0:
            raise RuntimeError(
                f"Last line of file '{fn}' read during metadata section."
            )
        elif line.strip() == "":
            break
        elif line.strip().startswith("Sequence name"):
            metadata["sequence"] = line.split(":,")[1]
        elif line.strip().startswith("Description"):
            metadata["description"] = line.split(":,")[1]
        elif line.strip().startswith("Acquired by"):
            metadata["username"] = line.split(":,")[1]
        elif line.strip().startswith("Data path"):
            metadata["datafile"] = line.split(":,")[1]
        elif line.strip().startswith("Report version"):
            metadata["version"] = int(line.split(":,")[1])

    if metadata.get("version", None) is None:
        raise RuntimeError(f"Report version in file '{fn}' was not specified.")

    samples = {}
    while len(lines) > 0:
        line = lines.pop(0)
        if len(lines) == 0:
            raise RuntimeError(f"Last line of file '{fn}' read during samples section.")
        elif line.strip() == "":
            break
        elif "Line#" in line:
            headers = [i.strip() for i in line.split(",")]
        else:
            data = [i.strip() for i in line.split(",")]
            sample = {
                "location": data[headers.index("Location")],
                "injection date": data[headers.index("Injection Date")],
                "acquisition": {
                    "method": data[headers.index("Acq Method Name")],
                    "version": data[headers.index("Acq Method Version")],
                },
                "integration": {
                    "method": data[headers.index("Injection DA Method Name")],
                    "version": data[headers.index("Injection DA Method Version")],
                },
                "offset": data[headers.index("Time offset")],
            }
            if sample["offset"] != "":
                sn = data[headers.index("Sample Name")]
                samples[sn] = sample

    svals = samples.values()
    if len(svals) == 0:
        raise RuntimeError(
            f"No complete sample data found in file '{fn}'. "
            "Have you added time offsets?"
        )
    r = next(iter(svals))
    # check that acquisition and integration methods are consistent throughout file:
    if any([s["acquisition"]["method"] != r["acquisition"]["method"] for s in svals]):
        logger.warning("Acquisition method is inconsistent in file '%s'.", fn)
    if any([s["integration"]["method"] != r["integration"]["method"] for s in svals]):
        logger.warning("Integration method is inconsistent in file '%s'.", fn)

    metadata["method"] = r["acquisition"]["method"]

    while len(lines) > 0:
        line = lines.pop(0)
        if len(lines) == 0:
            break
        elif line.strip() == "":
            break
        elif "Line#" in line:
            headers = [i.strip() for i in line.split(",")]
        else:
            data = [i.strip() for i in line.split(",")]
            sn = data[headers.index("Sample Name")]
            cn = data[headers.index("Compound")]

            h = data[headers.index("Peak Height")]
            if h != "":
                if "height" not in samples[sn]:
                    samples[sn]["height"] = {}
                n, s = tuple_fromstr(h)
                samples[sn]["height"][cn] = {"n": n, "s": s, "u": " "}

            A = data[headers.index("Area")]
            if A != "":
                if "area" not in samples[sn]:
                    samples[sn]["area"] = {}
                n, s = tuple_fromstr(A)
                samples[sn]["area"][cn] = {"n": n, "s": s, "u": " "}

            if metadata["version"] == 2:
                c = data[headers.index("Concentration")]
                u = "mmol/l"
            else:
                logger.warning(
                    "Report version '%d' in file '%s' not understood.",
                    metadata["version"],
                    fn,
                )
                c = data[headers.index("Concentration")]
                u = "mmol/l"
            if c != "":
                if "concentration" not in samples[sn]:
                    samples[sn]["concentration"] = {}
                n, s = tuple_fromstr(c)
                samples[sn]["concentration"][cn] = {"n": n, "s": s, "u": u}

            rt = data[headers.index("RT [min]")]
            if rt != "":
                if "retention time" not in samples[sn]:
                    samples[sn]["retention time"] = {}
                n, s = tuple_fromstr(rt)
                samples[sn]["retention time"][cn] = {"n": n, "s": s, "u": "min"}

    data = []
    for k, v in samples.items():
        # Remove unnecessary parameters
        del v["acquisition"]
        del v["integration"]
        v["sampleid"] = k
        # Process offset to uts
        offset = v.pop("offset")
        t = None
        for fmt in {"%H:%M:%S"}:
            try:
                t = datetime.datetime.strptime(offset, fmt)
            except ValueError:
                continue
        if t is None:
            try:
                td = datetime.timedelta(minutes=float(offset))
            except ValueError:
                raise RuntimeError(
                    f"It was not possible to parse offset '{offset}' present in file "
                    f"'{fn}' using known formats."
                )
        else:
            td = datetime.timedelta(hours=t.hour, minutes=t.minute, seconds=t.second)
        uts = td.total_seconds()
        point = {
            "uts": uts,
            "fn": fn,
            "raw": v,
        }
        data.append(point)
    return data, {"params": metadata}, False



import logging
import json
from pydantic import BaseModel
from . import drycal
from ... import dgutils

logger = logging.getLogger(__name__)


def process(
    fn: str,
    encoding: str = "utf-8",
    timezone: str = "localtime",
    parameters: BaseModel = None,
) -> tuple[list, dict, bool]:
    """
    Flow meter data processor

    This parser processes flow meter data.

    Parameters
    ----------
    fn
        File to process

    encoding
        Encoding of ``fn``, by default "utf-8".

    timezone
        A string description of the timezone. Default is "localtime".

    parameters
        Parameters for :class:`~dgbowl_schemas.yadg.dataschema_4_2.step.FlowData`.

    Returns
    -------
    (data, metadata, fulldate) : tuple[list, dict, bool]
        Tuple containing the timesteps, metadata, and full date tag. Whether full date
        is returned depends on the file parser.

    """
    if parameters.calfile is not None:
        dgutils.helpers.deprecated("parameters.calfile")
        with open(parameters.calfile, "r") as infile:
            calib = json.load(infile)
    else:
        calib = {}
    if parameters.convert is not None:
        dgutils.helpers.deprecated("parameters.convert")
        calib.update(parameters.convert)

    metadata = {}

    if parameters.filetype.startswith("drycal"):
        fulldate = False

        if parameters.filetype.endswith(".rtf") or fn.endswith("rtf"):
            ts, meta = drycal.rtf(fn, encoding, timezone, calib)
        elif parameters.filetype.endswith(".csv") or fn.endswith("csv"):
            ts, meta = drycal.sep(fn, ",", encoding, timezone, calib)
        elif parameters.filetype.endswith(".txt") or fn.endswith("txt"):
            ts, meta = drycal.sep(fn, "\t", encoding, timezone, calib)

        # check timestamps are increasing:
        warn = True
        ndays = 0
        for i in range(1, len(ts)):
            if ts[i]["uts"] < ts[i - 1]["uts"]:
                if warn:
                    logger.warning("DryCal log crossing day boundary. Adding offset.")
                    warn = False
                uts = ts[i]["uts"] + ndays * 86400
                while uts < ts[i - 1]["uts"]:
                    ndays += 1
                    uts = ts[i]["uts"] + ndays * 86400
                ts[i]["uts"] = uts

    metadata.update(meta)

    return ts, metadata, fulldate



"""
This parser handles the reading and processing of the legacy log files created by 
the LabView interface for the MCPT instrument. With a provided calibration, this 
parser calculates the temperature, inlet composition, and the inlet flow of the 
MCPT instrument.

.. admonition:: DEPRECATED in ``yadg-4.0``

    As of ``yadg-4.0``, this parser is deprecated and should not be used for new data.
    Please consider switching to the :mod:`~yadg.parsers.basiccsv` parser.

Usage
`````
Available since ``yadg-3.0``. Deprecated since ``yadg-4.0``. The parser supports the 
following parameters:

.. _yadg.parsers.meascsv.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_4_2.step.MeasCSV.Params

.. _parsers_meascsv_provides:

Provides
````````
The parser is used to extract all of the tabular data in the input file, storing 
them in the same format as :mod:`~yadg.parsers.basiccsv`, using the column headers 
as keys. The functionality exposed by this parser is using :mod:`~yadg.parsers.basiccsv` 
behind the scenes.


"""

from .main import process



"""
This parser handles the reading and processing of any tabular files, as long as 
the first line contains the column headers. By default, the second should contain
the units. The columns of the table must be separated using a separator such as 
``,``, ``;``, or ``\\t``. 

.. warning::

  Since ``yadg-4.2``, the parser handles sparse tables (i.e. tables with missing 
  data) by creating sparse `datagrams`, which means that the each element of the 
  header might not be present in each timestep.

.. note::

  :mod:`~yadg.parsers.basiccsv` attempts to deduce the timestamp from the column 
  headers, using :func:`yadg.dgutils.dateutils.infer_timestamp_from`. Alternatively, 
  the column(s) containing the timestamp data and their format can be provided using 
  parameters.

Usage
`````
Available since ``yadg-4.0``. The parser supports the following parameters:

.. _yadg.parsers.basiccsv.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_4_2.step.BasicCSV.Params

.. admonition:: DEPRECATED in ``yadg-4.2``

    The ``sigma``,  ``convert`` and ``calfile`` parameters are deprecated as of 
    ``yadg-4.2`` and will stop working in ``yadg-5.0``.

Provides
````````
The primary functionality of :mod:`~yadg.parsers.basiccsv` is to load the tabular 
data, and determine the Unix timestamp. The headers of the tabular data are taken 
`verbatim` from the file, and appear as ``raw`` data keys:

.. code-block:: yaml

  - uts: !!float
    fn:  !!str
    raw:
        "{{ column_name }}": 
            {n: !!float, s: !!float, u: !!str}


"""

from .main import process



"""
This module handles the reading of post-processed chromatography data, i.e.
files containing peak areas, concentrations, or mole fractions.

:mod:`~yadg.parsers.chromdata` loads the processed chromatographic data from the 
specified file, including the peak heights, areas, retention times, as well as the 
concentrations and mole fractions (normalised, unitless concentrations).

.. note:: 

  To parse trace data as present in raw chromatograms, use the 
  :mod:`~yadg.parsers.chromtrace` parser.

Usage
`````
Available since ``yadg-4.2``. The parser supports the following parameters:

.. _yadg.parsers.chromdata.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_4_2.step.ChromData.Params

.. _yadg.parsers.chromdata.formats:

Formats
```````
The ``filetypes`` currently supported by the parser are:

 - Inficon Fusion JSON format (``fusion.json``):  
   see :mod:`~yadg.parsers.chromdata.fusionjson`
 - Inficon Fusion zip archive (``fusion.zip``): 
   see :mod:`~yadg.parsers.chromdata.fusionzip`
 - Inficon Fusion csv export (``fusion.csv``): 
   see :mod:`~yadg.parsers.chromdata.fusioncsv`
 - Empa's Agilent LC csv export (``empalc.csv``): 
   see :mod:`~yadg.parsers.chromdata.empalccsv`
 - Empa's Agilent LC excel export (``empalc.xlsx``): 
   see :mod:`~yadg.parsers.chromdata.empalcxlsx`

.. _yadg.parsers.chromdata.provides:

Provides
````````
This raw data is stored, for each timestep, using the following format:

.. code-block:: yaml

  - uts: !!float
    fn:  !!str
    raw:
      sampleid: !!str             # sample name or valve ID
      height:                     # heights of the peak maxima
        "{{ species_name }}": 
            {n: !!float, s: !!float, u: !!str}
      area:                       # integrated areas of the peaks
        "{{ species_name }}": 
            {n: !!float, s: !!float, u: !!str}
      concentration:
        "{{ species_name }}": 
            {n: !!float, s: !!float, u: !!str}
      xout:                       # mole fractions (normalised concentrations)
        "{{ species_name }}": 
            {n: !!float, s: !!float, u: " "}
      retention time:
        "{{ species_name }}":
            {n: !!float, s: !!float, u: " "}

.. note::

    The mole fractions in ``xout`` always sum up to unity. If there is more than
    one outlet stream, or if some analytes remain unidentified, the values in 
    ``xout`` will not be accurate.

"""
from .main import process



"""
This module handles the reading and processing of files containing electrochemical
data, including BioLogic's EC-Lab file formats. The basic function of the parser is to:

#. Read in the technique data and create timesteps.
#. Collect metadata, such as the measurement settings and the loops
   contained in a given file.
#. Collect data describing the technique parameter sequences.

Usage
`````
Available since ``yadg-4.0``. The parser supports the following parameters:

.. _yadg.parsers.electrochem.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_4_2.step.ElectroChem.Params

.. _yadg.parsers.electrochem.formats:

Formats
```````
The ``filetypes`` currently supported by the parser are:

 - EC-Lab raw data binary file and parameter settings (``eclab.mpr``),
   see :mod:`~yadg.parsers.electrochem.eclabmpr`
 - EC-Lab human-readable text export of data (``eclab.mpt``),
   see :mod:`~yadg.parsers.electrochem.eclabmpt`
 - tomato's structured json output (``tomato.json``),
   see :mod:`~yadg.parsers.electrochem.tomatojson`
  
Provides
````````
Most standard techniques write data that can be understood as a series of
timesteps, with a measurement of the potential of the working electrode ``Ewe``,
the current applied by the potentiostat ``I``, and if present, also the potential
of the counter electrode ``Ece``. Depending on the technique, these quantitites
may be recorded as averages, i.e. ``<Ewe>``, ``<Ece>``, and ``<I>``. Technique 
metadata, such as the ``cycle number`` and the name of the ``technique`` are also
stored in each timestep:

.. code-block:: yaml

    - fn   !!str
      uts  !!float
      raw:
        Ewe:                 # potential of the working electrode, might be <Ewe>
          {n: !!float, s: !!float, u: !!str}
        Ece:                 # potential of the counter electrode, might be <Ece>
          {n: !!float, s: !!float, u: !!str}
        I:                   # current, might be <I>
          {n: !!float, s: !!float, u: !!str}
        cycle number: !!int
        technique:    !!str 

For impedance spectroscopy techniques (PEIS, GEIS), the data is by default
transposed to be made of spectroscopy traces. The data is split into traces using 
the ``cycle number`` column, and each trace is cast into a single timestep. Each 
trace now corresponds to a spectroscopy scan, indexed by the technique name (PEIS 
or GEIS). The timestep takes the following format:

.. code-block:: yaml

    - fn   !!str
    - uts  !!float
    - raw:
        traces:
          "{{ technique }}":
            "{{ col1 }}":
                [!!int, ...]
            "{{ col2 }}":
                {n: [!!float, ...], s: [!!float, ...], u: !!str}

.. note::

  This transposing behaviour can be toggled off by setting the ``transpose`` 
  parameter to ``False``, see documentation of the 
  :class:`~dgbowl_schemas.yadg.dataschema_4_2.step.ElectroChem.Params` class.


"""
from .main import process



"""
**fusionjson**: Processing Inficon Fusion json data format (json).
------------------------------------------------------------------

This is a fairly detailed data format, including the traces, the calibration applied,
and also the integrated peak areas and other processed information, which are parsed
by this module.

.. note ::

    To parse the raw trace data, use the :mod:`~yadg.parsers.chromtrace` module.

.. warning ::

    The detectors in the json files are not necessarily in a consistent order. To
    avoid inconsistent parsing of species which appear in both detectors, the
    detector keys are sorted. **Species present in both detectors** will be 
    **overwritten by the last detector** in alphabetical order.

Exposed metadata:
`````````````````

.. code-block:: yaml

    params:
      method:   !!str
      username: None
      version:  !!str
      datafile: !!str

.. codeauthor:: Peter Kraus
"""
import json
import logging
from ...dgutils.dateutils import str_to_uts

logger = logging.getLogger(__name__)


def process(fn: str, encoding: str, timezone: str) -> tuple[list, dict]:
    """
    Fusion json format.

    One chromatogram per file with multiple traces, and pre-analysed results.
    Only a subset of the metadata is retained, including the method name,
    detector names, and information about assigned peaks.

    Parameters
    ----------
    fn
        Filename to process.

    encoding
        Encoding used to open the file.

    timezone
        Timezone information. This should be ``"localtime"``.

    Returns
    -------
    ([chrom], metadata, fulldate): tuple[list, dict, bool]
        Standard timesteps, metadata, and date tuple.
    """

    with open(fn, "r", encoding=encoding, errors="ignore") as infile:
        jsdata = json.load(infile)
    metadata = {
        "filetype": "fusion.json",
        "params": {
            "method": jsdata.get("methodName", "n/a"),
            "version": jsdata.get("softwareVersion", {}).get("version", None),
            "datafile": jsdata.get("sequence", {}).get("location", None),
            "username": None,
        },
    }
    chrom = {"fn": str(fn), "raw": {}}
    chrom["uts"] = str_to_uts(jsdata["runTimeStamp"], timezone=timezone)

    valve = jsdata.get("annotations", {}).get("valcoPosition", None)
    if valve is not None:
        chrom["raw"]["valve"] = valve

    sampleid = jsdata.get("annotations", {}).get("name", None)
    if sampleid is not None:
        chrom["raw"]["sampleid"] = sampleid

    raw = {
        "height": {},
        "area": {},
        "concentration": {},
        "xout": {},
        "retention time": {},
    }

    # sort detector keys to ensure alphabetic order for ID matching
    for detname in sorted(jsdata["detectors"].keys()):
        detdict = jsdata["detectors"][detname]
        if "analysis" in detdict:
            for peak in detdict["analysis"]["peaks"]:
                if "label" not in peak:
                    continue
                if "height" in peak:
                    h = {"n": float(peak["height"]), "s": 1.0, "u": " "}
                    raw["height"][peak["label"]] = h
                if "area" in peak:
                    A = {"n": float(peak["area"]), "s": 0.01, "u": " "}
                    raw["area"][peak["label"]] = A
                if "concentration" in peak:
                    c = {
                        "n": float(peak["concentration"]),
                        "s": float(peak["concentration"]) * 1e-3,
                        "u": "%",
                    }
                    raw["concentration"][peak["label"]] = c
                if "normalizedConcentration" in peak:
                    x = {
                        "n": float(peak["normalizedConcentration"]),
                        "s": float(peak["normalizedConcentration"]) * 1e-3,
                        "u": "%",
                    }
                    raw["xout"][peak["label"]] = x
                if "top" in peak:
                    rt = {
                        "n": float(peak["top"]),
                        "s": 0.01,
                        "u": "s",
                    }
                    raw["retention time"][peak["label"]] = rt
        else:
            logger.warning("'analysis' of chromatogram not present in file '%s'", fn)

    for k in {"height", "area", "concentration", "xout", "retention time"}:
        if raw[k] == 0:
            del raw[k]

    chrom["raw"].update(raw)
    return [chrom], metadata, True



"""
This module handles the parsing of raw traces present in chromatography files, 
whether the source is a liquid chromatograph (LC) or a gas chromatograph (GC). 
The basic function of the parser is to:

#. read in the raw data and create timestamped `traces`
#. collect `metadata` such as the method information, sample ID, etc.

:mod:`~yadg.parsers.chromtrace` loads the chromatographic data from the specified
file, determines the uncertainties of the signal (y-axis), and explicitly 
populates the points in the time axis (x-axis), when required. 

Usage
`````
Available since ``yadg-4.0``. The parser supports the following parameters:

.. _yadg.parsers.chromtrace.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_4_2.step.ChromTrace.Params

.. admonition:: DEPRECATED in ``yadg-4.2``

    The ``calfile``, ``detectors`` and ``species`` parameters are deprecated
    as of ``yadg-4.2`` and will stop working in ``yadg-5.0``.

.. _yadg.parsers.chromtrace.formats:

Formats
```````
The ``filetypes`` currently supported by the parser are:

 - EZ-Chrom ASCII export (``ezchrom.asc``): 
   see :mod:`~yadg.parsers.chromtrace.ezchromasc`
 - Agilent Chemstation Chromtab (``agilent.csv``): 
   see :mod:`~yadg.parsers.chromtrace.agilentcsv`
 - Agilent OpenLab binary signal (``agilent.ch``): 
   see :mod:`~yadg.parsers.chromtrace.agilentch`
 - Agilent OpenLab data archive (``agilent.dx``): 
   see :mod:`~yadg.parsers.chromtrace.agilentdx`
 - Inficon Fusion JSON format (``fusion.json``): 
   see :mod:`~yadg.parsers.chromtrace.fusionjson`
 - Inficon Fusion zip archive (``fusion.zip``): 
   see :mod:`~yadg.parsers.chromtrace.fusionzip`

.. _yadg.parsers.chromtrace.provides:

Provides
````````
The raw data is stored, for each timestep, using the following format:

.. code-block:: yaml

  - uts: !!float
    fn:  !!str
    raw:
      traces:
        "{{ trace_name }}":        # detector name from the raw data file
          id:               !!int  # detector id for matching with calibration data
          t:                       # time-axis units are always seconds
            {n: [!!float, ...], s: [!!float, ...], u: "s"} 
          y:                       # y-axis units are determined from raw file
            {n: [!!float, ...], s: [!!float, ...], u: !!str}  

.. note::

  To parse processed data in the raw data files, such as integrated peak areas or 
  concentrations, use the :mod:`~yadg.parsers.chromdata` parser instead.

.. admonition:: DEPRECATED in ``yadg-4.2``

  The below functionality has been deprecated in ``yadg-4.2`` and will stop working
  in ``yadg-5.0``.

The data processing performed by :mod:`~yadg.parsers.chromtrace` is enabled 
automatically when calibration information is provided. The resulting data is stored 
in the ``derived`` entry in each `timestep`, and contains the following information:

.. code-block:: yaml

  - derived:
      peaks:
        "{{ trace_name }}":     # detector name from raw data file
          "{{ species_name }}": # species name matched from calibration
            peak:
              max:      !!int   # index of peak maximum
              llim:     !!int   # index of peak left limit
              rlim:     !!int   # index of peak right limit
            A:                  # integrated peak area 
              {n: !!float, s: !!float, u: !!str} 
            h:                  # height of peak maximum
              {n: !!float, s: !!float, u: !!str} 
            c:                  # calibrated "concentration" or other quantity
              {n: !!float, s: !!float, u: !!str} 
      height:                   # baseline-corrected height of peak maximum
        "{{ species_name }}":
            {n: !!float, s: !!float, u: !!str}
      area:                     # integrated area of the sample peak
        "{{ species_name }}":
            {n: !!float, s: !!float, u: !!str}
      concentration:            # concentration of species derived from area
        "{{ species_name }}":
            {n: !!float, s: !!float, u: !!str}
      xout:                     # normalised mol fractions of species
        "{{ species_name }}":
            {n: !!float, s: !!float, u: !!str}

.. note::

    The specification of dictionaries that ought to be passed to ``species`` and 
    ``detectors`` (or stored as json in ``"calfile"``) is described in 
    :func:`yadg.parsers.chromtrace.main.parse_detector_spec`. 

.. note::

    The quantity ``c``, determined for each integrated peak, may not necessarily 
    be concentration. It can also be mole fraction, as it is determined from the 
    peak area in ``A`` and any provided calibration specification. The calibration 
    interface allows for units to be supplied.

.. note::

    The mol fractions in ``xout`` always sum up to unity. If there is more than
    one outlet stream, these mol fractions have to be weighted by the flow rate 
    in a post-processing routine.

"""
from .main import process



"""
**eclabmpt**: Processing of BioLogic's EC-Lab ASCII export files.
-----------------------------------------------------------------

``.mpt`` files are made up of a header portion (with the technique
parameter sequences and an optional loops section) and a tab-separated
data table.

A list of techniques supported by this parser is shown in `the techniques table 
<yadg.parsers.electrochem.eclabmpr.techniques>`_.

File Structure of ``.mpt`` Files
````````````````````````````````

These human-readable files are sectioned into headerlines and datalines.
The header part of the ``.mpt`` files is made up of information that can be found 
in the settings, log and loop modules of the binary ``.mpr`` file.

If no header is present, the timestamps will instead be calculated from
the file's ``mtime()``.


Metadata
````````
The metadata will contain the information from the header of the file. 

.. note ::

    The mapping between metadata parameters between ``.mpr`` and ``.mpt`` files
    is not yet complete.

.. codeauthor:: Nicolas Vetsch <vetschnicolas@gmail.com>
"""
import re
import logging
import locale
from collections import defaultdict
from ...dgutils.dateutils import str_to_uts
from .eclabtechniques import get_resolution, technique_params, param_from_key

logger = logging.getLogger(__name__)

# Maps EC-Lab's "canonical" column names to yadg name and unit.
column_units = {
    '"Ri"/Ohm': ("'Ri'", "Ω"),
    "-Im(Z)/Ohm": ("-Im(Z)", "Ω"),
    "-Im(Zce)/Ohm": ("-Im(Zce)", "Ω"),
    "-Im(Zwe-ce)/Ohm": ("-Im(Zwe-ce)", "Ω"),
    "(Q-Qo)/C": ("(Q-Qo)", "C"),
    "(Q-Qo)/mA.h": ("(Q-Qo)", "mA·h"),
    "<Ece>/V": ("<Ece>", "V"),
    "<Ewe>/V": ("<Ewe>", "V"),
    "<I>/mA": ("<I>", "mA"),
    "|Ece|/V": ("|Ece|", "V"),
    "|Energy|/W.h": ("|Energy|", "W·h"),
    "|Ewe|/V": ("|Ewe|", "V"),
    "|I|/A": ("|I|", "A"),
    "|Y|/Ohm-1": ("|Y|", "S"),
    "|Z|/Ohm": ("|Z|", "Ω"),
    "|Zce|/Ohm": ("|Zce|", "Ω"),
    "|Zwe-ce|/Ohm": ("|Zwe-ce|", "Ω"),
    "Analog IN 1/V": ("Analog IN 1", "V"),
    "Analog IN 2/V": ("Analog IN 2", "V"),
    "Capacitance charge/µF": ("Capacitance charge", "µF"),
    "Capacitance discharge/µF": ("Capacitance discharge", "µF"),
    "Capacity/mA.h": ("Capacity", "mA·h"),
    "charge time/s": ("charge time", "s"),
    "Conductivity/S.cm-1": ("Conductivity", "S/cm"),
    "control changes": ("control changes", None),
    "control/mA": ("control_I", "mA"),
    "control/V": ("control_V", "V"),
    "control/V/mA": ("control_V/I", "V/mA"),
    "counter inc.": ("counter inc.", None),
    "Cp-2/µF-2": ("Cp⁻²", "µF⁻²"),
    "Cp/µF": ("Cp", "µF"),
    "Cs-2/µF-2": ("Cs⁻²", "µF⁻²"),
    "Cs/µF": ("Cs", "µF"),
    "cycle number": ("cycle number", None),
    "cycle time/s": ("cycle time", "s"),
    "d(Q-Qo)/dE/mA.h/V": ("d(Q-Qo)/dE", "mA·h/V"),
    "dI/dt/mA/s": ("dI/dt", "mA/s"),
    "discharge time/s": ("discharge time", "s"),
    "dQ/C": ("dQ", "C"),
    "dq/mA.h": ("dq", "mA·h"),
    "dQ/mA.h": ("dQ", "mA·h"),
    "Ece/V": ("Ece", "V"),
    "Ecell/V": ("Ecell", "V"),
    "Efficiency/%": ("Efficiency", "%"),
    "Energy charge/W.h": ("Energy charge", "W·h"),
    "Energy discharge/W.h": ("Energy discharge", "W·h"),
    "Energy/W.h": ("Energy", "W·h"),
    "error": ("error", None),
    "Ewe-Ece/V": ("Ewe-Ece", "V"),
    "Ewe/V": ("Ewe", "V"),
    "freq/Hz": ("freq", "Hz"),
    "half cycle": ("half cycle", None),
    "I Range": ("I Range", None),
    "I/mA": ("I", "mA"),
    "Im(Y)/Ohm-1": ("Im(Y)", "S"),
    "mode": ("mode", None),
    "Ns changes": ("Ns changes", None),
    "Ns": ("Ns", None),
    "NSD Ewe/%": ("NSD Ewe", "%"),
    "NSD I/%": ("NSD I", "%"),
    "NSR Ewe/%": ("NSR Ewe", "%"),
    "NSR I/%": ("NSR I", "%"),
    "ox/red": ("ox/red", None),
    "P/W": ("P", "W"),
    "Phase(Y)/deg": ("Phase(Y)", "deg"),
    "Phase(Z)/deg": ("Phase(Z)", "deg"),
    "Phase(Zce)/deg": ("Phase(Zce)", "deg"),
    "Phase(Zwe-ce)/deg": ("Phase(Zwe-ce)", "deg"),
    "Q charge/discharge/mA.h": ("Q charge/discharge", "mA·h"),
    "Q charge/mA.h": ("Q charge", "mA·h"),
    "Q charge/mA.h/g": ("Q charge", "mA·h/g"),
    "Q discharge/mA.h": ("Q discharge", "mA·h"),
    "Q discharge/mA.h/g": ("Q discharge", "mA·h/g"),
    "R/Ohm": ("R", "Ω"),
    "Rcmp/Ohm": ("Rcmp", "Ω"),
    "Re(Y)/Ohm-1": ("Re(Y)", "S"),
    "Re(Z)/Ohm": ("Re(Z)", "Ω"),
    "Re(Zce)/Ohm": ("Re(Zce)", "Ω"),
    "Re(Zwe-ce)/Ohm": ("Re(Zwe-ce)", "Ω"),
    "step time/s": ("step time", "s"),
    "THD Ewe/%": ("THD Ewe", "%"),
    "THD I/%": ("THD I", "%"),
    "time/s": ("time", "s"),
    "x": ("x", None),
    "z cycle": ("z cycle", None),
}


def _process_header(lines: list[str], timezone: str) -> tuple[dict, list, dict]:
    """Processes the header lines.

    Parameters
    ----------
    lines
        The header lines, starting at line 3 (which is an empty line),
        right after the `"Nb header lines : "` line.

    Returns
    -------
    tuple[dict, dict]
        A dictionary containing the settings (and the technique
        parameters) and a dictionary containing the loop indexes.

    """
    sections = "\n".join(lines).split("\n\n")
    # Can happen that no settings are present but just a loops section.
    assert not sections[1].startswith("Number of loops : "), "no settings present"
    # Again, we need the acquisition time to get timestamped data.
    assert len(sections) >= 3, "no settings present"
    technique = sections[1].strip()
    settings_lines = sections[2].split("\n")
    technique, params_keys = technique_params(technique, settings_lines)
    params = settings_lines[-len(params_keys) :]

    # Get the locale
    old_loc = locale.getlocale(category=locale.LC_NUMERIC)
    ewe_ctrl_re = re.compile(r"Ewe ctrl range : min = (?P<min>.+), max = (?P<max>.+)")
    ewe_ctrl_match = ewe_ctrl_re.search("\n".join(settings_lines))
    for loc in [old_loc, "de_DE.UTF-8", "en_GB.UTF-8", "en_US.UTF-8"]:
        try:
            locale.setlocale(locale.LC_NUMERIC, locale=loc)
            locale.atof(ewe_ctrl_match["min"].split()[0])
            locale.atof(ewe_ctrl_match["max"].split()[0])
            logging.debug(f"The locale of the current file is '{loc}'.")
            break
        except ValueError:
            logging.debug(f"Could not parse Ewe ctrl using locale '{loc}'.")

    # The sequence param columns are always allocated 20 characters.
    n_sequences = int(len(params[0]) / 20)
    params_values = []
    for seq in range(1, n_sequences):
        values = []
        for param in params:
            try:
                val = locale.atof(param[seq * 20 : (seq + 1) * 20])
            except ValueError:
                val = param[seq * 20 : (seq + 1) * 20].strip()
            values.append(val)
        params_values.append(values)
    params = [dict(zip(params_keys, values)) for values in params_values]
    settings_lines = [line.strip() for line in settings_lines[: -len(params_keys)]]

    # Parse the acquisition timestamp.
    timestamp_re = re.compile(r"Acquisition started on : (?P<val>.+)")
    timestamp_match = timestamp_re.search("\n".join(settings_lines))
    timestamp = timestamp_match["val"]
    for format in ("%m/%d/%Y %H:%M:%S", "%m.%d.%Y %H:%M:%S", "%m/%d/%Y %H:%M:%S.%f"):
        uts = str_to_uts(timestamp, format, timezone, False)
        if uts is not None:
            break
    if uts is None:
        raise NotImplementedError(f"Time format for {timestamp} not implemented.")

    loops = None
    if len(sections) >= 4 and sections[-1].startswith("Number of loops : "):
        # The header contains a loops section.
        loops_lines = sections[-1].split("\n")
        n_loops = int(loops_lines[0].split(":")[-1])
        indexes = []
        for n in range(n_loops):
            index = loops_lines[n + 1].split("to")[0].split()[-1]
            indexes.append(int(index))
        loops = {"n_loops": n_loops, "indexes": indexes}
    settings = {
        "posix_timestamp": uts,
        "locale": loc,
        "technique": technique,
        "raw": "\n".join(lines),
    }
    return settings, params, loops


def _process_data(lines: list[str], Eranges: list[float], Iranges: list[float]) -> dict:
    """Processes the data lines.

    Parameters
    ----------
    lines
        The data lines, starting right after the last header section.
        The first line is an empty line, the column names can be found
        on the second line.

    Returns
    -------
    dict
        A dictionary containing the datapoints in the format
        ([{column -> value}, ..., {column -> value}]). If the column
        unit is set to None, the value is an int. Otherwise, the value
        is a dict with value ("n"), sigma ("s"), and unit ("u").

    """
    # At this point the first two lines have already been read.
    # Remove extra column due to an extra tab in .mpt file column names.
    names = lines[1].split("\t")[:-1]
    columns, units = zip(*[column_units[n] for n in names])
    data_lines = lines[2:]
    datapoints = []
    for line in data_lines:
        values = line.split("\t")
        datapoint = {}
        for col, val, unit in list(zip(columns, values, units)):
            if unit is None:
                ival = int(locale.atof(val))
                if col == "I Range":
                    datapoint[col] = param_from_key("I_range", ival)
                    Irange = param_from_key("I_range", ival, to_str=False)
                else:
                    datapoint[col] = ival
        if "Ns" in datapoint:
            Erange = Eranges[datapoint["Ns"]]
        else:
            logger.info(
                "'Ns' is not in data table, using the first E range from 'params'."
            )
            Erange = Eranges[0]
        if "I Range" not in datapoint:
            logger.info(
                "'I Range' is not in data table, using the I range from 'params'."
            )
            if "Ns" in datapoint:
                Irstr = Iranges[datapoint["Ns"]]
            else:
                Irstr = Iranges[0]
            Irange = param_from_key("I_range", Irstr, to_str=False)
        for col, val, unit in list(zip(columns, values, units)):
            try:
                val = locale.atof(val)
            except ValueError:
                val = val.strip()
            if unit is None:
                continue
            assert isinstance(val, float), "`n` should not be string"
            s = get_resolution(col, val, unit, Erange, Irange)
            datapoint[col] = {"n": val, "s": s, "u": unit}
        datapoints.append(datapoint)
    return datapoints


def process(
    fn: str,
    encoding: str = "windows-1252",
    timezone: str = "UTC",
    transpose: bool = True,
) -> tuple[list, dict, bool]:
    """Processes EC-Lab human-readable text export files.

    Parameters
    ----------
    fn
        The file containing the data to parse.

    encoding
        Encoding of ``fn``, by default "windows-1252".

    timezone
        A string description of the timezone. Default is "UTC".

    Returns
    -------
    (data, metadata, fulldate) : tuple[list, dict, bool]
        Tuple containing the timesteps, metadata, and the full date tag. For mpt files,
        the full date might not be specified if header is not present.

    """
    file_magic = "EC-Lab ASCII FILE\n"
    with open(fn, "r", encoding=encoding) as mpt_file:
        assert mpt_file.read(len(file_magic)) == file_magic, "invalid file magic"
        mpt = mpt_file.read()
    lines = mpt.split("\n")
    nb_header_lines = int(lines[0].split()[-1])
    header_lines = lines[: nb_header_lines - 3]
    data_lines = lines[nb_header_lines - 3 :]
    settings, params, loops = {}, [], {}
    # Store current LC_NUMERIC before we do anything:
    old_loc = locale.getlocale(category=locale.LC_NUMERIC)
    if nb_header_lines <= 3:
        logger.warning("Header contains no settings and hence no timestamp.")
        start_time = 0.0
        fulldate = False
        Eranges = [20.0]
        Iranges = ["Auto"]
    else:
        settings, params, loops = _process_header(header_lines, timezone)
        start_time = settings.get("posix_timestamp")
        fulldate = True
        Eranges = []
        Iranges = []
        for el in params:
            E_range_max = el.get("E_range_max", float("inf"))
            E_range_min = el.get("E_range_min", float("-inf"))
            Eranges.append(E_range_max - E_range_min)
            Iranges.append(el.get("I_range", "Auto"))
    data = _process_data(data_lines, Eranges, Iranges)
    # Arrange all the data into the correct format.
    # TODO: Metadata could be handled in a nicer way.
    metadata = {"settings": settings, "params": params}
    timesteps = []
    # If the technique is an impedance spectroscopy, split it into
    # traces at different cycle numbers and put each trace into its own timestep
    if settings["technique"] in {"PEIS", "GEIS"} and transpose:
        # Grouping by cycle.
        cycles = defaultdict(list)
        for d in data:
            cycles[d["cycle number"]].append(d)
        # Casting cycles into traces.
        for ti, td in cycles.items():
            trace = {col: [d[col] for d in td] for col in td[0].keys()}
            for key, val in trace.items():
                if not isinstance(val[0], dict):
                    continue
                trace[key] = {k: [i[k] for i in val] for k in val[0]}
                # Reducing unit list to just a string.
                trace[key]["u"] = set(trace[key]["u"]).pop()
            uts = start_time + trace["time"]["n"][0]
            trace["time"]["n"] = [i - trace["time"]["n"][0] for i in trace["time"]["n"]]
            timesteps.append(
                {
                    "uts": uts,
                    "fn": fn,
                    "raw": {"traces": {settings["technique"]: trace}},
                }
            )
        return timesteps, metadata, fulldate
    # All other techniques have multiple timesteps.
    for d in data:
        uts = start_time + d["time"]["n"]
        d["technique"] = settings["technique"]
        timesteps.append({"fn": fn, "uts": uts, "raw": d})
    # reset to original LC_NUMERIC
    locale.setlocale(category=locale.LC_NUMERIC, locale=old_loc)
    return timesteps, metadata, fulldate
